from typing import Callable, Iterable
from .core import ParaO
from .action import Act, RecursiveAction


class RunAction(RecursiveAction):
    func: Callable[[ParaO], None]  # TODO: add dynamic tasks

    def _func(self, depth: int, inner: Iterable[Act]): ...  # TODO


class Task(ParaO):
    @RecursiveAction
    def remove(self, depth: int, inner: Iterable[Act]):
        pass

    @RecursiveAction
    def status(self, depth: int, inner: Iterable[Act]):
        pass



from typing import Any
from unittest import TestCase
from unittest.mock import ANY, Mock, call

from parao.action import (
    Action,
    MissingParameterValueOrOverride,
    Plan,
    RecursiveAction,
    ValueAction,
)
from parao.core import ParaO, Param, UntypedParameter


class TestAction(TestCase):
    def test_action(self):
        func = Mock()

        class Foo(ParaO):
            act = Action[Any, None, []](func)

        sentinel = object()

        with Plan().use(run=True):
            foo = Foo({"act": sentinel})
        func.assert_called_once_with(foo, sentinel)

        func.reset_mock()
        foo = Foo()
        foo.act()
        func.assert_called_once_with(foo)

    def test_value_action(self):
        mock = Mock()

        class Foo(ParaO):
            @ValueAction
            def act(self, value: int):
                return mock(self, value)

            @ValueAction
            def act2(self, value):
                return (123, value)

        with self.assertRaises(MissingParameterValueOrOverride):
            Foo().act()

        with Plan().use(run=True), self.assertWarns(UntypedParameter):
            foo = Foo({"act": 123})
        mock.assert_called_once_with(foo, 123)

        mock.reset_mock()
        foo.act()
        mock.assert_called_once_with(foo, 123)

        mock.reset_mock()
        mock.return_value = 456
        self.assertEqual(foo.act(321), 456)
        mock.assert_called_once_with(foo, 321)

        sentinel = object()
        self.assertEqual(foo.act2(sentinel), (123, sentinel))

    def test_recursive_action(self):
        mock = Mock(return_value=None)

        class FooBase(ParaO):
            act = RecursiveAction(mock)

        class Foo1(FooBase):
            pass

        class Foo2(FooBase):
            foo1 = Param[Foo1]()

        class Foo3(FooBase):
            foo2 = Param[Foo2]()

        with Plan().use(run=True):
            foo3 = Foo3(act=True)
        mock.assert_has_calls(
            [call(foo3, 0, ANY), call(foo3.foo2, 1, ANY), call(foo3.foo2.foo1, 2, ANY)]
        )

        mock.reset_mock()
        foo3.act(1)
        mock.assert_has_calls([call(foo3, 0, ANY), call(foo3.foo2, 1, ())])

        mock.reset_mock()
        (foo3 := Foo3()).act()
        mock.assert_has_calls(
            [call(foo3, 0, ANY), call(foo3.foo2, 1, ANY), call(foo3.foo2.foo1, 2, ANY)]
        )

        mock.reset_mock()
        (foo3 := Foo3({(Foo1, "act"): False})).act()
        mock.assert_has_calls([call(foo3, 0, ANY), call(foo3.foo2, 1, ANY)])

        mock.reset_mock()
        (foo3 := Foo3()).act(1)
        mock.assert_has_calls([call(foo3, 0, ANY), call(foo3.foo2, 1, ())])

        mock.reset_mock()
        (foo3 := Foo3()).act(0)
        mock.assert_has_calls([call(foo3, 0, ())])

        # self.assertEqual(foo3.act, ())



from collections import defaultdict
from dataclasses import dataclass
from functools import cached_property, lru_cache, partial
from itertools import count
from math import inf
from types import GenericAlias
from typing import (
    Any,
    Callable,
    Collection,
    Generator,
    Iterable,
    Mapping,
    Protocol,
    Self,
    TypeVar,
    _GenericAlias,
    get_type_hints,
)
from warnings import warn
from os.path import dirname

from .cast import cast, Opaque
from .misc import ContextValue, safe_len, safe_repr

__all__ = ["UNSET", "ParaO", "Param", "Prop", "Const"]
_warn_skip = (dirname(__file__),)


class Unset(Opaque): ...


UNSET = Unset()

Unset.__new__ = lambda _: UNSET

_param_counter = count()


# type KeyE = type | object | str
type KeyE = str | type | AbstractParam
type KeyT = tuple[KeyE, ...]
type KeyTE = KeyT | KeyE
type TypT = type | GenericAlias
type PrioT = int | float
type Mapish[K, V] = Mapping[K, V] | Iterable[tuple[K, V]]


@dataclass(frozen=True, slots=True)
class Arg:
    key: KeyT
    val: Any
    prio: PrioT = 0
    offset: int = 0

    def __repr__(self):
        parts = list(map(repr, self.key[self.offset :]))
        parts.append(f"val={self.val!r}")
        if self.prio:
            parts.append(f"prio={self.prio!r}")
        return f"{self.__class__.__name__}({', '.join(parts)})"

    __hash__ = object.__hash__  # value can be unhashable!

    @property
    def is_final(self):
        return self.offset >= len(self.key)

    @property
    def effective_key(self):
        return self.key[self.offset :]

    def __gt__(self, other: "Arg | None") -> bool:
        return other is None or self.prio > other.prio

    def __lt__(self, other: "Arg | None") -> bool:
        return other is not None and self.prio < other.prio

    # @lru_cache
    def _solve_value(self, ref: "ParaOMeta"):
        last = len(self.key) - 1
        off = self.offset
        # match owner class filters
        while (
            isinstance((key := self.key[off]), type)
            and off < last
            and issubclass(ref, key)
        ):
            off += 1

        op = ref.__own_parameters__
        if key in op.vals or (key := op.get(key)):
            return (
                key,
                Arg(self.key, self.val, self.prio, off + 1),
                off == self.offset,  # is_final & no class filter
            )
        else:
            return None, None, False

    @lru_cache
    def solve_class[T: type](self, cls: T) -> tuple[PrioT, "T | Arg | Arguments"]:
        last = len(self.key) - 1
        off = self.offset
        # match owner class filters
        while (
            off <= last
            and isinstance((key := self.key[off]), type)
            and issubclass(cls, key)
        ):
            off += 1
        if (off == last and key == "__class__") or off > last:
            if not isinstance(self.val, (type, Arg, Arguments)):
                raise TypeError(f"{self!r} resolved to non-class")
            return self.prio, self.val
        return 0, UNSET

    def get_root_of(self, arg: "Arg") -> "Arg | None":
        if (
            self.key is arg.key
            and self.val is arg.val
            and self.prio is arg.prio
            and self.offset <= arg.offset
        ):
            return self


class Arguments(tuple["Arguments | Arg", ...]):
    is_final = False

    @classmethod
    def make(cls, *args: "Arguments | HasArguments | dict[KeyTE, Any]", **kwargs: Any):
        return cls._make(args + (kwargs,)) if kwargs else cls._make(args)

    @classmethod
    def _make(cls, args: "tuple[Arguments | HasArguments | dict[KeyTE, Any], ...]"):
        sub = []
        if arg := cls._ctxargs():
            sub.append(arg)

        for arg in args:
            arg = getattr(arg, "__args__", arg)
            if isinstance(arg, cls):
                if arg:
                    sub.append(arg)
            elif isinstance(arg, dict):
                if arg:
                    sub.append(cls.from_dict(arg.items()))
            else:
                raise TypeError(f"unsupported argument type: {type(arg)}")

        return cls.from_list(sub)

    @classmethod
    def from_dict(
        cls,
        k2v: Mapping[KeyTE, Any] | Iterable[tuple[KeyTE, Any]],
        prio: PrioT = 0,
    ):
        if callable(items := getattr(k2v, "items", None)):
            k2v = items()
        return cls(Arg(k if isinstance(k, tuple) else (k,), v, prio) for k, v in k2v)

    @classmethod
    def from_list(cls, args: "list[Arguments | Arg]") -> "Arguments":
        """Turn an iterable into arguments. Avoids unnecessary nesting or repeated creation of empty Arguments."""
        match args:
            case []:
                return cls.EMPTY
            case [Arguments()]:
                return args[0]
        return cls(args)

    def __repr__(self):
        return self.__class__.__name__ + (tuple.__repr__(self) if self else "()")

    def solve_value(self, param, owner, name) -> tuple[Arg | None, "Arguments"]:
        com, val, sub = self._solve_values(owner)
        su = sub.get(param)

        return val.get(param), com if su is None else Arguments.from_list(su)

    @lru_cache
    def _solve_values(self, ref: "ParaOMeta"):
        com = []
        val: dict[AbstractParam, Arg] = {}
        sub: defaultdict[AbstractParam, list[Arguments | Arg]]

        if False:
            sub = defaultdict(com.copy)  # TODO: optimize this
        else:  # optimized behaviour

            @defaultdict
            def sub():
                nonlocal com
                if len(com) > 1:
                    com = [Arguments.from_list(com)]
                return com.copy()

        for arg in self:
            if isinstance(arg, Arg):
                k, d, r = arg._solve_value(ref)
                if not r:
                    sub.get(k, com).append(arg)
                if d is not None:
                    assert k is not None
                    if d.is_final:
                        if isinstance(d.val, (Arguments, Arg)):
                            sub[k].append(d.val)
                        else:
                            val[k] = max(d, val.get(k))
                    else:
                        sub[k].append(d)
            else:
                c, v, s = arg._solve_values(ref)
                val.update(v)
                for k, v in s.items():
                    sub[k].extend(v)
                com.append(c)  # must be done after filling sub

        return Arguments.from_list(com), val, sub

    @lru_cache
    def solve_class(self, cls):
        prio = -inf

        for v in self:
            while isinstance(v, (Arguments, Arg)):
                p, v = v.solve_class(cls)
                if v is UNSET:
                    break
                elif prio <= p:
                    prio = p
                    cls = v

        return prio, cls

    def get_root_of(self, arg: "Arg") -> "Arg | None":
        for a in self:
            if r := a.get_root_of(arg):
                return r


Arguments.EMPTY = Arguments()
Arguments._ctxargs = ContextValue("ContextArguments", default=Arguments.EMPTY)


class HasArguments(Protocol):
    __args__: Arguments


eager = ContextValue[bool]("eager", default=False)


class _OwnParameters(dict[str, "AbstractParam"]):
    __slots__ = "vals"
    vals: set["AbstractParam"]

    def __init__(self, cls: "ParaOMeta"):
        super().__init__(
            (name, param)
            for name in dir(cls)
            if not name.startswith("__")
            and isinstance((param := getattr(cls, name)), AbstractParam)
        )
        self.vals = set(self.values())

    cache: dict["ParaOMeta", "_OwnParameters"] = {}


class ParaOMeta(type):
    @property
    def __own_parameters__(cls) -> _OwnParameters:
        if (val := _OwnParameters.cache.get(cls)) is None:
            val = _OwnParameters.cache[cls] = _OwnParameters(cls)
        return val

    def __setattr__(cls, name, value):
        if not name.startswith("__"):
            if cache := _OwnParameters.cache.get(cls):
                if old := cache.get(name):
                    old.__set_name__(cls, None)
                    cache.vals.remove(old)
                    del cache[name]
            if isinstance(value, AbstractParam):
                value.__set_name__(cls, name)
                if cache:
                    cache.vals.add(value)
                    cache[name] = value
        return super().__setattr__(name, value)

    def __delattr__(cls, name):
        if not name.startswith("__"):
            if cache := _OwnParameters.cache.get(cls):
                if old := cache.get(name):
                    old.__set_name__(cls, None)
                    cache.vals.remove(old)
                    del cache[name]
        return super().__delattr__(name)

    def __cast_from__(cls, value, original_type):
        if value is UNSET:
            return cls()
        if isinstance(value, cls):
            return value
        return cls(value)

    def __call__(
        cls, *args: Arguments | HasArguments | dict[KeyTE, Any], **kwargs: Any
    ) -> Self:
        arg = Arguments._make(args + (kwargs,) if kwargs else args)
        ret = cls.__new__(arg.solve_class(cls)[1])
        ret.__args__ = arg
        ret.__init__()
        if eager():
            for name, param in ret.__class__.__own_parameters__.items():
                if param.eager:
                    getattr(ret, name)
        return ret


class ParaO(metaclass=ParaOMeta):
    __args__: Arguments  # | UNSET

    @cached_property
    def __inner__(self) -> tuple["ParaO", ...]:
        ret = []
        for name, param in self.__class__.__own_parameters__.items():
            if param.significant:
                val = getattr(self, name)
                if isinstance(val, ParaO):
                    ret.append(val)
                elif isinstance(val, Expansion):
                    ret.extend(val.expand())
                elif isinstance(val, (dict, list, tuple, set, frozenset)):
                    queue = [iter((val,))]
                    while queue:
                        for curr in queue[-1]:
                            if isinstance(curr, (list, tuple, set, frozenset)):
                                queue.append(iter(curr))
                                break
                            elif isinstance(curr, dict):
                                queue.append(iter(curr.keys()))
                                queue.append(iter(curr.values()))
                                break
                            elif isinstance(curr, ParaO):
                                ret.append(curr)
                        else:
                            queue.pop()
        return tuple(ret)


class TypedAlias(GenericAlias):
    def __call__(self, *args, **kwds):
        kwds.setdefault("type", self.__args__[0])
        return super().__call__(*args, **kwds)

    @classmethod
    def convert(cls, ga: _GenericAlias):
        return cls(ga.__origin__, ga.__args__)

    @classmethod
    def init_subclass(cls, other: type):
        if "type" in other.__dict__:
            return
        for ob in other.__orig_bases__:
            if isinstance(ob, cls):
                typ = ob.__args__[0]
                if isinstance(typ, TypeVar):
                    if other.__type_params__[:1] != ob.__args__[:1]:
                        raise TypeError(f"first typevar of {other} must be {typ}")
                else:
                    setattr(other, "type", typ)
                return


class UntypedParameter(RuntimeWarning): ...


type ExpansionFilter = Collection[KeyE | Collection[KeyE]] | Callable[
    [Expansion, ParaO], bool
]


class DuplicateParameter(RuntimeError): ...


@lru_cache
def _solve_name(param: "Param", icls: "ParaOMeta") -> str | None:
    lut = param._owner2name
    for cls in icls.__mro__:
        if cls in lut:
            return lut[cls]


@lru_cache
def _get_type_hints(cls: "ParaOMeta"):
    return get_type_hints(cls)


### actual code
class AbstractParam[T]:
    significant: bool = True

    def __class_getitem__(cls, key):
        return TypedAlias.convert(super().__class_getitem__(key))

    def __init_subclass__(cls):
        super().__init_subclass__()
        TypedAlias.init_subclass(cls)

    __slots__ = ("__dict__", "_owner2name", "_id")

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)
        self._owner2name = {}
        self._id = next(_param_counter)

    def __set_name__(self, cls, name):
        if name:
            old = self._owner2name.get(cls, None)
            if old is not None:
                raise DuplicateParameter(f"{self} on {cls} with names: {old}, {name}")
            self._owner2name[cls] = name
        else:
            del self._owner2name[cls]
        _solve_name.cache_clear()

    def _solve_name(self, instance: "ParaO"):
        return _solve_name(self, type(instance))

    def _solve_type(self, cls: "ParaOMeta", name: str):
        return _get_type_hints(cls).get(name, UNSET)

    def _cast(self, val, typ):
        if typ is not UNSET:
            try:
                exp = cast(val, Expansion[typ])
            except TypeError:
                pass
            else:
                if isinstance(exp, Expansion):
                    raise exp
        return cast(val, typ)

    def _get(self, val: Any, name: str, instance: "ParaO") -> T:
        typ = self.type
        if typ is UNSET:
            typ = self._solve_type(type(instance), name)
        if typ is UNSET:
            warn(
                f"{type(self)} {name} on {type(instance)}",
                category=UntypedParameter,
                skip_file_prefixes=_warn_skip,
            )
            return val
        return self._cast(val, typ)

    def __get(
        self,
        arg: Arg | None,
        name: str,
        instance: "ParaO",
        *args: Arguments | Arg,
    ):
        val = UNSET if arg is None or arg.prio < self.min_prio else arg.val

        try:
            with Arguments._ctxargs(Arguments.from_list(args)):
                return self._get(val, name, instance)
        except Expansion as exp:
            exp.process(self, instance, arg)
            exp._get = partial(self.__get, arg, name, instance, *args)
            return exp
        except Exception as exc:
            exc.add_note(f"parameter {name}={safe_repr(self)} on {safe_repr(instance)}")
            raise

    def __get__(self, instance: "ParaO", owner: type | None = None):
        if instance is None:
            return self
        name = self._solve_name(instance)

        arg, sub = instance.__args__.solve_value(self, type(instance), name)

        instance.__dict__[name] = val = self.__get(arg, name, instance, sub)
        return val

    min_prio: float = -inf
    eager: bool = True
    collect: ExpansionFilter = ()
    type: type | Unset
    type = UNSET


class Const[T](AbstractParam[T]):
    def __init__(self, value: T, **kwargs):
        super().__init__(value=value, **kwargs)

    def __get__(self, instance, owner=None) -> T:
        return self.value


class AbstractDecoParam[T, F: Callable](AbstractParam[T]):
    def __init__(self, func: F, **kwargs):
        assert callable(func)
        super().__init__(func=func, **kwargs)

    def __new__(cls, func: F | None = None, **kwargs):
        if func is None:
            return partial(cls, **kwargs)
        return super().__new__(cls)

    func: F


class Prop[T](AbstractDecoParam[T, Callable[[ParaO], T]]):
    def _solve_type(self, cls, name):
        typ = self.func.__annotations__.get("return", UNSET)
        if typ is UNSET:
            typ = super()._solve_type(cls, name)
        return typ

    def _get(self, val, name, instance) -> T:
        val = super()._get(val, name, instance)
        return self.func(instance) if val is UNSET else val


class MissingParameterValue(TypeError): ...


class Param[T](AbstractParam[T]):
    def __init__(self, default=UNSET, **kwargs):
        super().__init__(default=default, **kwargs)

    def _get(self, val, name, instance):
        val = super()._get(val, name, instance)
        if val is UNSET:
            if self.default is UNSET:
                raise MissingParameterValue(name)
            else:
                return self.default
        return val


class ExpansionGeneratedKeyMissingParameter(RuntimeWarning): ...


class Expansion[T](BaseException):
    @classmethod
    def __cast_from__(cls, value, original_type):
        (typ,) = original_type.__args__
        result = cast(value, tuple[typ, ...])
        if isinstance(result, tuple):
            return cls(result)
        else:
            raise NotImplementedError

    def __init__(self, values: Iterable[T]):
        super().__init__()
        assert iter(values)  # ensure values is iterable
        self.values = values
        self._frames: list[tuple[ParaOMeta, Param, Arg | None]] = []

    _get: Callable[[Arg], T | Self]

    def test(self, item: KeyE | Iterable[KeyE]):
        match item:
            case AbstractParam():
                return self.param is item
            case str():
                return self.param_name == item
            case type():
                return isinstance(self.source, item)
            case _:
                return all(map(self.test, item))

    def process(self, param: AbstractParam, inst: ParaO, arg: Arg):
        # leave marks of origin
        if not self._frames:
            self.param = param
            self.source = inst
            # do we need these? or are they only for _unwind construction
            self.arg = arg
        # is it collected here?
        if param.collect and (
            param.collect(self, inst)
            if callable(param.collect)
            else any(map(self.test, param.collect))
        ):
            return  # this will add ._get
        # keep track of key to dial-down to the origin
        self._frames.append(
            (inst.__class__, param, inst.__args__.get_root_of(self.arg))
        )
        raise  # self # but don't to avoid mangling the traceback

    def make_key(
        self,
        use_arg: bool = True,
        *,
        dont: Collection[KeyE] = (),
        use_cls: bool = True,
        use_param: bool = True,
        use_name: bool = True,
        want: Collection[KeyE] | None = None,
    ):
        dont = set(dont)
        dont.difference_update(
            dont_cls := tuple(d for d in dont if isinstance(d, type))
        )

        if want is not None:
            want = set(want)
            want.difference_update(
                want_cls := tuple(w for w in want if isinstance(w, type))
            )

        rkey = []
        for cls, param, root in self._frames:
            if use_arg and root is not None:
                rkey = list(root.effective_key[::-1])
                continue

            name = _solve_name(param, cls)

            if (
                use_param
                and param not in dont
                and name not in dont
                and (want is None or param in want or name in want)
            ):
                rkey.append(name if use_name else param)
            if (
                use_cls
                and not issubclass(cls, dont_cls)
                and (want is None or issubclass(cls, want_cls))
            ):
                if not rkey:
                    rkey.append(name if use_name else param)
                    warn(
                        f"force added {rkey[0]!r}",
                        ExpansionGeneratedKeyMissingParameter,
                    )
                rkey.append(cls)
        return tuple(reversed(rkey))

    def expand(self, prio: PrioT = 0, **kwargs) -> Generator[T]:
        key = self.make_key(**kwargs)
        for val in self.values:
            res = self._get(Arg(key, val, prio))
            if isinstance(res, Expansion):
                yield from res.expand(prio=prio, **kwargs)
            else:
                yield res

    @staticmethod
    def generate(typ: ParaOMeta, args: Arguments, **kwargs) -> Generator[ParaO]:
        try:
            yield typ(args)
        except Expansion as exp:
            exp._get = lambda arg: typ(Arguments((args, arg)))
            try:
                yield from exp.expand(**kwargs)
            except Exception as exc:
                exc.add_note(f"while expanding: {exp!r}")
                raise

    @property
    def param_name(self):
        return self.param._solve_name(self.source)

    def __repr__(self):
        parts = [f"< {safe_len(self.values)} values >"]
        if self._frames:
            parts.append(f"param={self.param}")
            parts.append(f"source={self.source}")
        return f"{self.__class__.__name__}({', '.join(parts)})"



from binascii import b2a_hex
from collections import OrderedDict
from functools import lru_cache
from hashlib import sha256
from operator import attrgetter
from types import FunctionType, ModuleType
from typing import Any, Hashable, Protocol
from warnings import warn

primitives = (str, int, float, type(None), bytes, complex, type(Ellipsis))

importables = (type, FunctionType)  # TODO: add more of the builtin stuff

try:
    from numpy import dtype, ndarray
except ModuleNotFoundError:

    class ndarray: ...  # phony placeholder

else:
    primitives += (dtype,)


primitives_set = frozenset(primitives)


class UsesLocalsError(ValueError): ...


class UsesMainWarning(RuntimeWarning): ...


class UsesPickleWarning(RuntimeWarning): ...


class UnsupportedError(ValueError): ...


class _SHash(list):
    end = sha256(b"").digest()

    def __call__(self, value: Any) -> bytes:
        if value.__class__ in primitives_set:
            return _repr(value)
        if isinstance(value, (type, FunctionType, ModuleType)):
            return _imp(value)

        if value in self:
            return sha256(str(len(self) - self.index(value)).encode()).digest()

        try:
            self.append(value)

            if shash := getattr(value, "__shash__", None):
                return shash(self)

            ret = sha256(_imp(value.__class__))

            if isinstance(value, ndarray):
                inner = value.tolist()
            elif isinstance(value, dict):
                inner = map(self.item, value.items())
                if not isinstance(value, OrderedDict):
                    inner = sorted(inner)
            elif isinstance(value, (list, tuple)):
                inner = map(self, value)
            elif isinstance(value, (set, frozenset)):
                inner = sorted(map(self, value))
            elif isinstance(value, (range, slice)):
                inner = map(self, attrgetter("start", "stop", "step")(value))
            elif isinstance(value, primitives):
                return _repr(value)
            elif callable(red := getattr(value, "__reduce__", None)):
                warn(UsesPickleWarning(value.__class__))
                ret.update(self(red()))
                return ret.digest()
            else:
                raise UnsupportedError(value)

            for val in inner:
                ret.update(self(val))
            ret.update(self.end)
            return ret.digest()

        finally:
            self.pop()

    def item(self, item: tuple[Hashable, Any]) -> bytes:
        h = sha256(self(item[0]))
        h.update(self(item[1]))
        return h.digest()


class SHashable(Protocol):
    def __shash__(self, shash: _SHash) -> bytes: ...


def _repr(value) -> bytes:
    return sha256(repr(value).encode()).digest()


@lru_cache
def _imp(value: type | FunctionType) -> bytes:
    if isinstance(value, ModuleType):
        rep = value.__name__
    else:
        rep = f"{value.__module__}:{value.__qualname__}"
    if "<locals>" in rep:
        raise UsesLocalsError(value)
    if rep.startswith("__main__"):
        warn(UsesMainWarning(value))
    return sha256(rep.encode()).digest()


def bin_hash(value: Any) -> bytes:
    return _SHash()(value)


def hex_hash(value: Any) -> bytes:
    return b2a_hex(bin_hash(value))



from operator import attrgetter
from unittest import TestCase
from unittest.mock import Mock
from parao.core import (
    UNSET,
    AbstractParam,
    Arg,
    Arguments,
    Const,
    Expansion,
    ParaO,
    Param,
    MissingParameterValue,
    Prop,
    Unset,
    UntypedParameter,
    eager,
    DuplicateParameter,
    ExpansionGeneratedKeyMissingParameter,
)


uniq_object = object()


class TestUnset(TestCase):
    def test(self):
        self.assertIsInstance(UNSET, Unset)
        self.assertIs(UNSET, Unset())
        self.assertIs(Unset(), Unset())


class TestArg(TestCase):

    def test_create(self):
        key = "foo", "bar", "boo"
        a = Arg(key, uniq_object)
        self.assertEqual(a.key, key)
        self.assertEqual(a.val, uniq_object)
        self.assertEqual(a.prio, 0)
        self.assertEqual(a.offset, 0)
        # prio argument
        self.assertEqual(Arg(key, uniq_object, 123).prio, 123)
        # offset argument
        self.assertEqual(Arg(key, uniq_object, 0, 123).offset, 123)

    def test_repr(self):

        self.assertEqual(
            repr(Arg(("foo", "bar"), val=123, prio=321)),
            "Arg('foo', 'bar', val=123, prio=321)",
        )
        self.assertEqual(
            repr(Arg(("foo", "bar"), val=123, offset=1)),
            "Arg('bar', val=123)",
        )


class TestArguments(TestCase):

    def test_create(self):
        tpl = (1, "foo", uniq_object)  # the are actually bad types ...
        self.assertTupleEqual(Arguments(tpl), tpl)

        self.assertEqual(
            Arguments(
                [
                    Arg(("foo",), uniq_object, prio=123),
                    Arg(("foo", "bar"), uniq_object, prio=123),
                ]
            ),
            Arguments.from_dict(
                {"foo": uniq_object, ("foo", "bar"): uniq_object},
                prio=123,
            ),
        )
        self.assertEqual(
            Arguments(
                [
                    Arg(("foo",), uniq_object, prio=123),
                    Arg(("foo", "bar"), uniq_object, prio=123),
                ]
            ),
            Arguments.from_dict(
                [
                    ("foo", uniq_object),
                    (("foo", "bar"), uniq_object),
                ],
                prio=123,
            ),
        )

        self.assertIs(Arguments.from_list([]), Arguments.EMPTY)
        self.assertIs(Arguments.from_list([a := Arguments()]), a)
        self.assertEqual(
            Arguments.from_list([Arg(("foo",), uniq_object)]),
            Arguments([Arg(("foo",), uniq_object)]),
        )

        self.assertRaises(TypeError, lambda: ParaO(123))

    def test_repr(self):

        self.assertEqual(repr(Arguments.make({})), "Arguments()")
        self.assertEqual(
            repr(Arguments.make(key=123)),
            "Arguments(Arg('key', val=123),)",
        )
        self.assertEqual(
            repr(Arguments.make(foo=123, bar=456)),
            "Arguments(Arg('foo', val=123), Arg('bar', val=456))",
        )


class TestParam(TestCase):
    def test_param(self):
        self.assertIs(Param(type=(o := object())).type, o)
        self.assertIs(Param[o := object()]().type, o)
        self.assertRaises(TypeError, lambda: Param[int, str])

    def test_typed_alias(self):
        class Sentinel:
            pass

        class StrangeParam(AbstractParam):
            type = Sentinel

        self.assertIs(StrangeParam().type, Sentinel)

        with self.assertRaises(TypeError):

            class BadTypeVarParam[Q, T](AbstractParam[T]):
                pass

    def test_specialized(self):
        uniq_const = object()
        uniq_aux = object()
        uniq_return = object()
        uniq_override = object()

        class Special(ParaO):
            const = Const(uniq_const)

            @Prop(aux=uniq_aux)
            def prop(self) -> object:
                return uniq_return

        self.assertIs(Special(const=None).const, uniq_const)
        self.assertIs(Special.prop.aux, uniq_aux)
        self.assertIs(Special().prop, uniq_return)
        self.assertIs(Special(prop=uniq_override).prop, uniq_override)


class TestParaO(TestCase):
    def test_create(self):
        ParaO()

        class Sub(ParaO): ...

        self.assertIsInstance(Sub(), Sub)
        self.assertIsInstance(ParaO({ParaO: Sub}), Sub)
        self.assertIsInstance(ParaO({"__class__": Sub}), Sub)

        self.assertRaises(TypeError, lambda: ParaO({ParaO: 123}))

        with self.assertRaises(DuplicateParameter):
            Sub.foo1 = Sub.foo2 = Param()

    def test_own_params(self):
        class Sub(ParaO):
            foo: int = Param()
            bar: str = Param()

        self.assertEqual(Sub.__own_parameters__, {"foo": Sub.foo, "bar": Sub.bar})

        Sub.boo = Param(type=float)

        self.assertEqual(Sub.__own_parameters__["boo"], Sub.boo)

        Sub.boo = Param(type=complex)

        self.assertEqual(Sub.__own_parameters__["boo"], Sub.boo)

        del Sub.foo, Sub.bar, Sub.boo

        self.assertEqual(Sub.__own_parameters__, {})

    def test_resolution_simple(self):

        class Sub(ParaO):
            foo: int = Param()
            bar = Param(None, type=str)
            boo = Param[bool]()
            bad = Param(None)

        self.assertEqual(Sub.boo.type, bool)

        with self.assertRaises(MissingParameterValue):
            Sub().foo
        with self.assertWarns(UntypedParameter):
            Sub().bad

        self.assertEqual(Sub({"foo": 123}).foo, 123)
        self.assertEqual(Sub({Sub.foo: 123}).foo, 123)
        self.assertEqual(Sub({(Sub, "foo"): 123}).foo, 123)
        self.assertEqual(Sub({(Sub, Sub, "foo"): 123}).foo, 123)
        self.assertEqual(Sub({(Sub, Sub.foo): 123}).foo, 123)

        self.assertEqual(Sub().bar, None)
        self.assertEqual(Sub(bar=123).bar, "123")

    def test_resolution_complex(self):

        class Sub(ParaO):
            foo: int = Param()
            bar: str = Param(None)

        class Sub2(Sub):
            boo: bool = Param()

        self.assertEqual(Sub({Sub: Sub2, "boo": True}).boo, True)

        class Wrap(ParaO):
            one: Sub = Param()
            other: Sub2 = Param()

        class More(ParaO):
            inner: Wrap = Param()

        for addr in [("one", "bar"), (Wrap.one, Sub.bar), (Wrap.one, Sub, "bar")]:
            with self.subTest(addr=addr):
                self.assertEqual(Wrap({addr: 123}).one.bar, "123")
                self.assertEqual(Wrap({addr: 123}).other.bar, None)

        # providing a dict
        self.assertEqual(Wrap(one=dict(foo=123)).one.foo, 123)

        # unsing instance's args
        self.assertEqual(Wrap(one=Sub2(foo=123)).one.foo, 123)
        self.assertEqual(Wrap(one=Sub2(foo=123).__args__).one.foo, 123)

        # direct instance providing
        self.assertEqual(Wrap(one=Sub(foo=123)).one.foo, 123)
        self.assertIs(Wrap(one=(s := Sub())).one, s)

        # self.assertEqual(More().inner)

        # obj = Wrap({(Sub, "foo"): 123})
        # self.assertEqual(obj.one.foo, 123)
        # self.assertEqual(obj.other.foo, 123)

    def test_common_base(self):
        class Base(ParaO):
            foo = Param[int](0)

        class Ext1(Base):
            pass

        class Ext2(Base):
            ext1 = Param[Ext1]()

        ext2 = Ext2(foo=1)
        self.assertEqual(ext2.foo, 1)
        self.assertEqual(ext2.ext1.foo, 0)

    def test_expansion(self):

        class Foo(ParaO):
            bar = Param[int]()

        with eager(False):
            f = Foo(bar=[1, 2, 3])
            # raises on access
            self.assertRaises(Expansion, lambda: f.bar)

        with eager(True):
            self.assertRaises(Expansion, lambda: Foo(bar=[1, 2, 3]))
            try:
                Foo(bar=[1, 2, 3])
            except Expansion as exp:
                self.assertEqual(exp.param, Foo.bar)
                self.assertEqual(exp.param_name, "bar")
                self.assertEqual(exp.values, (1, 2, 3))

        self.assertEqual(repr(Expansion([1, 2, 3])), "Expansion(< 3 values >)")

    def test_collect(self):

        class Foo(ParaO):
            bar = Param[int]()

        # function based
        func = Mock(return_value=True)

        class Wrap(ParaO):
            foo = Param[Foo](collect=func)

        with eager(True):
            inst = Wrap(bar=[1, 2, 3])
        exp = inst.foo
        func.assert_called_once_with(exp, inst)
        self.assertIsInstance(exp.source, Foo)
        self.assertIsInstance(exp, Expansion)
        # self.assertEqual(exp._unwind, [])
        self.assertEqual(exp.make_key(), ("bar",))
        self.assertEqual(exp.make_key(False), (Foo, "bar"))
        self.assertEqual(exp.make_key(False, use_cls=False), ("bar",))
        self.assertEqual(exp.make_key(False, use_name=False), (Foo, Foo.bar))
        with self.assertWarns(ExpansionGeneratedKeyMissingParameter):
            self.assertEqual(
                exp.make_key(False, use_param=False),
                (Foo, "bar"),
            )
        with self.assertWarns(ExpansionGeneratedKeyMissingParameter):
            self.assertEqual(
                exp.make_key(False, want=(Foo,), use_name=False), (Foo, Foo.bar)
            )
        self.assertEqual(exp.make_key(False, want=(Foo.bar,)), ("bar",))
        self.assertIsInstance(repr(exp), str)

        # bare argument based
        items = [[Foo], [Foo.bar], ["bar"]]
        for coll in items + [[it] for it in items]:
            with self.subTest(coll=coll), eager(True):
                Wrap.foo.collect = coll
                self.assertIsInstance(Wrap(bar=[1, 2, 3]).foo, Expansion)

    def test_expand(self):
        # uses two level expandable scenario

        class Foo(ParaO):
            bar = Param[int]()

        class Mid(ParaO):
            boo = Param[int](0)
            foo = Param[Foo]()

        class Wrap2(ParaO):
            mid = Param[Mid](collect=Mock(return_value=True))

        with eager(True):
            self.assertEqual(Wrap2(bar=[1, 2, 3]).mid.make_key(), ("bar",))
            self.assertEqual(
                Wrap2({("foo", "bar"): [1, 2, 3]}).mid.make_key(), ("foo", "bar")
            )
            self.assertEqual(
                Wrap2(foo=dict(bar=[1, 2, 3])).mid.make_key(), (Mid, "foo", "bar")
            )
            self.assertSequenceEqual(
                list(map(attrgetter("foo.bar"), Wrap2(bar=[1, 2, 3]).mid.expand())),
                [1, 2, 3],
            )
            self.assertSequenceEqual(
                list(
                    map(
                        attrgetter("foo.bar"),
                        Wrap2({("foo", "bar"): [1, 2, 3]}).mid.expand(),
                    )
                ),
                [1, 2, 3],
            )
            self.assertSequenceEqual(
                list(
                    map(
                        attrgetter("foo.bar"),
                        Wrap2({("mid", "foo", "bar"): [1, 2, 3]}).mid.expand(),
                    )
                ),
                [1, 2, 3],
            )
            self.assertSequenceEqual(
                list(
                    map(
                        attrgetter("boo", "foo.bar"),
                        Wrap2(boo=[1, -1], bar=[1, 2, 3]).mid.expand(),
                    )
                ),
                [
                    (1, 1),
                    (1, 2),
                    (1, 3),
                    (-1, 1),
                    (-1, 2),
                    (-1, 3),
                ],
            )

    def test_inner(self):
        class In(ParaO):
            exp = Param[int](0)

            @Prop
            def uniq(self) -> int:
                return id(self)

        class Out(ParaO):
            in1 = Param[In](collect=[In.exp])
            in2 = Param[list[In]]([])

            in3u = Param[In](significant=False)

            @Prop
            def in3(self) -> dict:
                return {
                    "deep": [
                        "nested",
                        ("structure", {"with", frozenset({"some", (self.in3u,) * 2})}),
                    ]
                }

        out1 = Out()
        self.assertEqual(out1.__inner__, (out1.in1, *out1.in2, out1.in3u, out1.in3u))

        out2 = Out(in2=[In(), In()])
        self.assertEqual(out2.__inner__, (out2.in1, *out2.in2, out2.in3u, out2.in3u))

        with eager(True):
            out3 = Out({("in1", "exp"): [1, 2]})
        self.assertEqual(out3.__inner__[0].exp, 1)
        self.assertEqual(out3.__inner__[1].exp, 2)
        self.assertEqual(out3.__inner__[2:], (out3.in3u, out3.in3u))



from contextlib import contextmanager
from functools import lru_cache, partial
from inspect import Parameter, signature
from typing import Any, Callable, Concatenate, Iterable, Type

from .core import (
    UNSET,
    AbstractDecoParam,
    Expansion,
    MissingParameterValue,
    ParaO,
    Unset,
    eager,
)
from .misc import ContextValue


@lru_cache
def _method_1st_arg_annotation[T](
    func: Callable[Concatenate[Any, T, ...], Any],
) -> Type[T] | Unset:
    for i, param in enumerate(signature(func).parameters.values()):
        if i == 1:
            if (
                param.kind
                in (
                    Parameter.POSITIONAL_ONLY,
                    Parameter.POSITIONAL_OR_KEYWORD,
                    Parameter.VAR_POSITIONAL,
                )
                and param.annotation is not Parameter.empty
            ):
                return param.annotation
            break
    return UNSET


class Act[T, R](partial):
    func: "Action[T, R]"
    args: tuple[ParaO, T]


class Action[T, R, **Ps](AbstractDecoParam[T, Callable[Concatenate[ParaO, Ps], R]]):
    significant = False

    def _get(self, val, name, instance) -> Act[T, R]:
        return Plan.add(self, instance, super()._get(val, name, instance))

    __get__: Callable[..., Act[T, R] | Expansion]

    def __call__(
        self, inst: ParaO, value: T | Unset, *args: Ps.args, **kwargs: Ps.kwargs
    ):
        if value is UNSET:
            return self.func(inst, *args, **kwargs)
        else:
            return self.func(inst, value, *args, **kwargs)


class MissingParameterValueOrOverride(MissingParameterValue): ...


class ValueAction[T, R](Action[T, R, T]):
    def _solve_type(self, cls, name):
        return _method_1st_arg_annotation(self.func)

    __get__: Callable[..., Act[T, R] | Expansion]

    def __call__(self, inst: ParaO, value: T | Unset, override: T | Unset = UNSET):
        if override is not UNSET:
            value = override
        if value is UNSET:
            raise MissingParameterValueOrOverride(self._solve_name(inst))
        return self.func(inst, value)


class RecursiveAction(Action[int | bool | None, None, [int, Iterable[Act]]]):
    func: Callable[[ParaO, int, Iterable[Act]], Iterable[Act] | None]
    _func: Callable[[ParaO, int, Iterable[Act]], Iterable[Act] | None] | None = None

    def _solve_inner(self, inst: ParaO) -> Iterable[Act]:
        name = self._solve_name(inst)
        cls = type(self)
        for inner in inst.__inner__:
            if other := inner.__class__.__own_parameters__.get(name):
                if isinstance(other, cls):
                    yield getattr(inner, name)

    def __call__(
        self,
        inst: ParaO,
        value: int | bool | Unset = UNSET,
        override: int | bool | Unset = UNSET,
        *,
        outer: int = True,
        depth: int = 0,
    ):
        if override is not UNSET:
            value = override
        if value is UNSET:
            value = outer
        elif value is False or value < 0:
            return

        inner = self._solve_inner(inst) if value else ()

        _func = self._func or self.func
        ret = _func(inst, depth, inner)
        if ret is None:
            ret = inner

        kwargs = dict(depth=depth + 1)
        if value is not True and value > 0:
            kwargs["outer"] = value - 1

        for r in ret:
            r(**kwargs)


class Plan(list[Act]):
    current = ContextValue["Plan"]("currentActionPlan", default=None)

    @classmethod
    def add(cls, action: Action, inst: ParaO, value: Any | Unset) -> Act:
        act = Act(action, inst, value)
        if value is not UNSET and (curr := cls.current()) is not None:
            curr.append(act)
        return act

    @contextmanager
    def use(self, /, run: bool = False):
        with self.current(self), eager(True):
            yield
            if run:
                self.run()

    def run(self):
        while self:
            self.pop(0)()



from binascii import b2a_hex
from collections import OrderedDict
import sys
from types import ModuleType
from unittest import TestCase
from unittest.mock import Mock, patch

from parao.shash import (
    _SHash,
    UnsupportedError,
    UsesLocalsError,
    UsesMainWarning,
    UsesPickleWarning,
    bin_hash,
    hex_hash,
)


def fake_main_func(): ...


fake_main_func.__module__ = "__main__"


class Unpickable:
    __reduce__ = None


list_in_self = []
list_in_self.append(list_in_self)

value2expected = [
    (1, b"6b86b273ff34fce19d6b804eff5a3f5747ada4eaa22f1d49c01e52ddb7875b4b"),
    (1.234, b"00a4e3d579394c0826980361ff3ca00e4915d0dc33b3bcb8775453aed9e8b16d"),
    (5.678j, b"2852241032ed1e4fba97014e64f3e1935f2a9eaa2093755590f851583bfb0377"),
    ("foo", b"af693da41462ec24753ce51403bd1ee5b500cd6941f62f3c97456dd8ffa7f952"),
    (b"bar", b"0f0cd6bc85f1855d43efd8f861ecd52a7a9360b7cd32537c85118e24d657abae"),
    (None, b"dc937b59892604f5a86ac96936cd7ff09e25f18ae6b758e8014a24c7fa039e91"),
    (Ellipsis, b"4637e99b28a1ce112e8f4009dbf144f3a75a04d3e4b0c30b084aef21c5e3cfe6"),
    ((1, 2, 3), b"e45859a8f662a552e3f3b3194e1e9957ac711422939adccb8d0821d299262336"),
    ([1, 2, 3], b"8e8be76c77fe4142b767006751320e33e9544becc298f50a64f5a5fc539a024a"),
    ({1, 2, 3}, b"13e72599b081ec70bc81b345e4a0c1b6051fb82cecd1bf826afea2e47a05f8d3"),
    ({1: 2, 3: 4}, b"b50bba3f366c0d96c8eb25148429c7040a004ac25d57cd131c63d25e5eeb3cb7"),
    (
        OrderedDict({1: 2, 3: 4}),
        b"a5fd8b9a41cdf8cf5d8def2972a3d5fe80c3e3a4eefb79db3f6984492b80ea48",
    ),
    (
        OrderedDict({3: 4, 1: 2}),
        b"cb2cb90289016eb180776fc056fa87d7037a0b91304817aa37b217187530c9ff",
    ),
    (
        range(1, 2, 3),
        b"8e19905aea3ca08a0edec0cc85a8e80eaae69193cb93136214c5f570a0021121",
    ),
    (
        slice(1, 2, 3),
        b"bf92450de58ece8b5525a1c442f882d94d5a9b0f1cbde7850510d0a74838cb01",
    ),
    # wonky types
    (
        sys,
        b"518b67e652531c5fe7e25d6b2c3b4ef6224e7d90da2091967dd47eb082b26a19",
    ),
    (
        list_in_self,
        b"e03f50c7b0624e5831ace9788342ec9d44cc68725cde30db6ac371adc470d1a3",
    ),
]


class ndarray:
    def tolist(self): ...


class Custom:
    def __shash__(self, enc: _SHash) -> bytes: ...


class TestShash(TestCase):
    def test_values(self):
        for value, expected in value2expected:
            with self.subTest(value=value, expected=expected):
                self.assertEqual(hex_hash(value), expected)

    def test_equivalence(self):
        self.assertEqual(b2a_hex(bin_hash(None)), hex_hash(None))

        self.assertEqual(hex_hash(set(range(100))), hex_hash(set(reversed(range(100)))))

        self.assertEqual(hex_hash({1: 2, 3: 4}), hex_hash({3: 4, 1: 2}))

    def test_difficulties(self):
        self.assertRaises(UnsupportedError, lambda: bin_hash(Unpickable()))

    def test_bad_unpickable(self):
        for probe in [
            object(),
            # from types
            len,
            [].append,
            object.__init__,
            object().__str__,
            str.join,
            # dict.__dict__["fromkeys"],
        ]:
            with self.subTest(probe=probe), self.assertWarns(UsesPickleWarning):
                bin_hash(probe)

    def test_locals(self):
        def make_closure():
            a = 1

            def func():
                return a  # pragma: no cover

            return func

        self.assertRaises(UsesLocalsError, lambda: bin_hash(make_closure()))

    def test_main(self):
        self.assertWarns(UsesMainWarning, lambda: bin_hash(fake_main_func))

    def test_numpy(self):
        numpy = Mock(spec=ModuleType)

        numpy.ndarray = ndarray
        numpy.dtype = Mock(spec=type)

        with patch.dict(sys.modules, numpy=numpy):
            del sys.modules["parao.shash"]

            from parao.shash import hex_hash

            nda = Mock(spec=ndarray)
            nda.tolist = Mock(return_value=[[[1], [2], [3]]])

            self.assertEqual(
                hex_hash(nda),
                b"6524a1a7d0e04156c52ae2000ba8081c769d66d381e06c06ec26fac508c7da75",
            )

            nda.tolist.assert_called_once()

    def test_custom(self):

        custom = Mock()
        custom.__shash__ = Mock()
        custom.__shash__.return_value = b"foobar"

        s = _SHash()

        self.assertEqual(s(custom), b"foobar")
        custom.__shash__.assert_called_once_with(s)



from .core import ParaO, Param, Const, Prop
from .cli import CLI
from .action import Action, ValueAction, RecursiveAction
from .cast import Opaque  # noqa: F401

__all__ = [
    "ParaO",
    "Param",
    "Const",
    "Prop",
    "CLI",
    "Action",
    "ValueAction",
    "RecursiveAction",
]



from contextlib import AbstractContextManager, contextmanager
from contextvars import ContextVar
from typing import Any, Iterable, overload

_sentinel = object()

__all__ = ["ContextValue", "context_manager_set", "safe_repr", "safe_len"]


@contextmanager
def context_manager_set[T](contextvar: ContextVar[T], value: T):
    token = contextvar.set(value)
    try:
        yield token.old_value
    finally:
        contextvar.reset(token)


class ContextValue[T]:
    __slots__ = ("contextvar",)

    def __init__(self, name: str, *, default: T = _sentinel):
        if default is _sentinel:
            self.contextvar = ContextVar[T](name)
        else:
            self.contextvar = ContextVar[T](name, default=default)

    @overload
    def __call__(self) -> T:
        """Return the current context variable value or the global default, if any."""

    @overload
    def __call__(self, *, default: T) -> T:
        """Return the current context variable value or the given default."""

    @overload
    def __call__(self, value: T) -> AbstractContextManager[T, None]:
        """Set the context variable value to the given value and returns a context manager that passes the previous value and reset the value upon exit."""

    def __call__(self, value: T = _sentinel, *, default: T = _sentinel):
        if value is _sentinel:
            if default is _sentinel:
                return self.contextvar.get()
            else:
                return self.contextvar.get(default)
        else:
            return context_manager_set(self.contextvar, value)


def safe_repr(obj: Any) -> str:
    """get representation, exception safe"""
    try:
        return repr(obj)
    except Exception:
        return object.__repr__(obj)


def safe_len[T](obj: Any, default: T = None) -> int | T:
    """get len of obj, return default on failure"""
    try:
        return len(obj)
    except Exception:
        return default


class PeekableIter[T]:
    __slots__ = (
        "_iter",
        "_head",
    )

    def __init__(self, it: Iterable[T]):
        self._iter = iter(it)
        self._head = _sentinel

    def __iter__(self):
        return self

    def __next__(self) -> T:
        if self._head is _sentinel:
            return next(self._iter)
        else:
            ret = self._head
            self._head = _sentinel
            return ret

    def peek(self, default=_sentinel) -> T:
        if self._head is _sentinel:
            try:
                self._head = next(self._iter)
            except StopIteration:
                if default is _sentinel:
                    raise
                else:
                    return default
        return self._head


def is_subseq(needles, haystack):
    haystack = iter(haystack)
    return all(needle in haystack for needle in needles)


# def exhaust(it: Iterable) -> None:
#     deque(it, maxlen=0)



from itertools import repeat
from types import UnionType, NoneType
from typing import _AnnotatedAlias, Any, Protocol, Union, get_args, get_origin

_numeric = int, float, complex


class Opaque:
    """Ignored during casting unless they are Castable."""


class Castable(Protocol):
    @classmethod
    def __cast_from__(cls, value, original_type): ...


def cast(val: Any, typ: type) -> Any:
    typ0 = typ  # the orignal type
    if isinstance(typ, _AnnotatedAlias):
        (typ,) = typ.__args__
    ori = get_origin(typ)

    if cast_to := getattr(val, "__cast_to__", None):
        try:
            return cast_to(typ, typ0)
        except NotImplementedError:
            pass

    if cast_from := getattr(ori or typ, "__cast_from__", None):
        try:
            return cast_from(val, typ0)
        except NotImplementedError:
            pass

    if isinstance(val, Opaque):
        return val

    if ori:
        args = get_args(typ)

        if ori is UnionType or ori is Union:
            for arg in args:
                try:
                    return cast(val, arg)
                except (TypeError, ValueError):
                    pass
            else:
                raise TypeError(f"could not cast {val!r} to {typ}")

        # container types
        if isinstance(ori, type):  # pragma: no branch

            if isinstance(val, (str, bytes)):
                raise TypeError(f"can't cast {type(val)} to container")

            if issubclass(ori, tuple):
                if not args:
                    if val:
                        raise TypeError("too many values given")
                    return val if isinstance(val, ori) else ori()
                if args[1:] == (Ellipsis,):
                    return ori(map(cast, val, repeat(args[0])))

                ret = ori(map(cast, val, args))
                if len(ret) != len(args):
                    raise TypeError("wrong number of values")
                return ret

            if issubclass(ori, (list, set, frozenset)):
                (typ1,) = args
                return ori(map(cast, val, repeat(typ1)))

            if issubclass(ori, dict):
                typK, typV = args
                return ori(
                    {
                        cast(k, typK): cast(v, typV)
                        for k, v in (val.items() if isinstance(val, dict) else val)
                    }
                )

        raise TypeError(f"type no understood: {typ}")
    elif typ is Any:
        return val
    # primitive types
    elif typ is None or typ is NoneType:
        if val is not None:
            raise TypeError("invalid value for None")
        return None
    elif isinstance(val, typ):
        return val
    elif isinstance(val, str) and (
        typ is bool or isinstance(type, typ) and issubclass(typ, bool)
    ):
        raise TypeError("can't cast {val!r} to {typ}")
    else:
        ret = typ(val)
        if isinstance(val, _numeric) and isinstance(ret, _numeric) and ret != val:
            raise ValueError(f"cast to {typ} not accurate: {val!r}!={ret!r}")
        return ret
